/**CFile***********************************************************************

  FileName    [ cceCmd.c ]

  PackageName [ cce ]

  Synopsis    [ Tcl commands ]

  Author      [ Stephen A. Edwards <sedwards@eecs.berkeley.edu> ]

  Copyright   [ ]

******************************************************************************/

#include "cceInt.h"
#include "tcl.h"
#include <stdio.h>

/**AutomaticStart*************************************************************/

/**AutomaticEnd***************************************************************/

/**Function********************************************************************

  Synopsis           [ Tcl command for reading a slot's contents ]

  Description        [ Find the named slot, write the result ]

  SideEffects        []

  CommandName        [ cce_slot ] 	   

  CommandSynopsis    [ Return the contents of an object's slot ]  

  CommandArguments   [ <i>object</i> <i>slotName</i> <i>?slotValue?</i> ]

  CommandDescription []  

******************************************************************************/
static int
SlotCmd( ClientData dummy, Tcl_Interp * interp, int argc, char * argv[] )
{
  int id;
  int slot;
  int type;
  Cce_Object object;
  Cce_Object class;
  Cce_Object slotObject;
  char buf[10];

  if ( argc != 3 && argc != 4 ) {
    Tcl_AppendResult( interp, "wrong # args: should be \"", argv[0],
		      " object slotName ?slotValue?\"", (char *) NULL );
    return TCL_ERROR;
  }

  if ( sscanf(argv[1],"obj%d", &id ) ) {
    if ( object = Cce_ObjectOfId( id ) ) {
      class = Cce_ObjectReadClass( object );
      if ( (slot = Cce_ClassFindNamedSlot( class, argv[2], &type )) >= 0 ) {
	
	if ( argc == 3 ) {

	  /*
	   * Read a slot's value
	   */

	  switch ( type ) {

	  case CCE_OBJECT:
	    slotObject = (Cce_Object) object[slot];
	    while ( slotObject ) {
	      sprintf( buf, "obj%d", Cce_ObjectReadId( slotObject ) );
	      Tcl_AppendElement( interp, buf );
	      slotObject = Cce_ObjectReadNext( slotObject );
	    }
	    return TCL_OK;
	    break;

	  case CCE_STRING:
	    Tcl_SetResult( interp, (char *) object[slot], TCL_STATIC );
	    return TCL_OK;
	    break;

	  case CCE_INTEGER:
	    sprintf( interp->result, "%d", (int) object[slot] );
	    return TCL_OK;
	    break;

	  default:
	    sprintf( interp->result, "Unknown slot type %d!", type );
	    return TCL_ERROR;
	    break;
	  }	  

	} else {

	  /*
	   * Write a slot's value
	   */

	  interp->result = "Writing currently unimplemented";
	  return TCL_ERROR;

	}

      } else {
	Tcl_AppendResult( interp, "unrecognized slot \"", argv[2],
			  "\" in class \"",
			  Cce_ObjectSlotReadString( class, CCE_CLASS_NAME ) ,
			  "\"", (char *) NULL );
	return TCL_ERROR;
      }      
    }    
  }
  Tcl_AppendResult( interp, "Unknown object identifier \"", argv[1],
		    "\". obj0 through obj", (char *) NULL );
  sprintf( buf, "%d", cce_numObjects - 1);
  Tcl_AppendResult( interp, buf, " are currently valid.", (char *) NULL );
  return TCL_ERROR;
}

/**Function********************************************************************

  Synopsis           [ Tcl command for parsing a file ]

  Description        [ Open the file, parse it, and return the object
  		       it created ]

  SideEffects        []

  CommandName        [ cce_parse_file ] 	   

  CommandSynopsis    [ Parse a file and return the object it creates ]  

  CommandArguments   [ <i>filename</i> ]

  CommandDescription []  

******************************************************************************/
static int
ParseFileCmd( ClientData dummy, Tcl_Interp * interp, int argc, char * argv[] )
{
  Cce_Object object;
  int result;
  char buf[20];
  Tcl_DString errorString;

  FILE * file;

  if ( argc != 2 ) {
    Tcl_AppendResult( interp, "wrong # args: should be \"", argv[0],
		      " filename\"", (char *) NULL );
    return TCL_ERROR;
  }

  file = fopen( argv[1], "r" );
  if ( !file ) {
    Tcl_AppendResult( interp, "unable to open source file \"", argv[1],
		      "\"", (char *) NULL );
    return TCL_ERROR;
  }

  object = 0;

  Tcl_DStringInit( &errorString );

  result = Cce_ParseFile( file, argv[1], &object, &errorString );
  fclose( file );

  if ( object ) {
    sprintf( buf, "obj%d", Cce_ObjectReadId( object ) );
  } else {
    buf[0] = '\0';
  }

  Tcl_AppendElement( interp, buf );
  Tcl_AppendElement( interp, Tcl_DStringValue( &errorString ) );
  Tcl_DStringFree( &errorString );

  return TCL_OK;
}


/**Function********************************************************************

  Synopsis           [ Initialize the cce package ]

  Description        [ ]

  SideEffects        [ Adds command to the interpreter ]

  SeeAlso            []

******************************************************************************/
int
Cce_Init( Tcl_Interp * interp )
{

#ifdef HAVE_PKG_PROVIDE
  int code;

  code = Tcl_PkgProvide( interp, "Cce", "1.0" );
  if ( code != TCL_OK ) {
    return code;
  }
#endif

  Cce_ObjectsInitialize();  

  Tcl_CreateCommand( interp, "cce_slot", SlotCmd, (ClientData) NULL,
		     (Tcl_CmdDeleteProc *) NULL );
  Tcl_CreateCommand( interp, "cce_parse_file", ParseFileCmd, (ClientData) NULL,
		     (Tcl_CmdDeleteProc *) NULL );

  return TCL_OK;
}


