%{ /* -*- C -*- */

  /*
   * Simplified bison grammar for C++ code
   *
   * Stephen Edwards
   */

#include "cceInt.h"

#define yylloc ccelloc

#define YYDEBUG 1

#define CCE_DEBUG_CLASSES 1
#define CCE_DEBUG_BASES 2
int debug = CCE_DEBUG_CLASSES | CCE_DEBUG_BASES;

static int status = 0;
static char * currentFilename;
static Tcl_DString * parseErrorString;

EXTERN int ccelineno;
EXTERN char * ccetext;

static Cce_Object toplevelObject;

#define SetLocation(o,l) \
  Cce_AstObjectWriteLine( o, (l).first_line ); \
  Cce_AstObjectWriteFilename( o, Cce_StringDup(currentFilename) )

%}

%union {
  char * string;
  Cce_Object object;
  int i;
}

%token COMMENT
%token CLASS STRUCT ENUM UNION FRIEND
%token OPERATOR TYPEDEF EXTERN_T DOUBLECOLON
%token <i> ACCESS QUALIFIER
%token <string> COMMENT COMMENTFIELD
%token <string> IDENTIFIER OP TOKEN

%type <string> optDefault defaultValue optIdentifier
%type <string> operator

%type <object> classDeclaration optBaseList baseList baseSpecifier
%type <object> classMemberList structMemberList
%type <object> accessMemberList optMemberList memberList
%type <object> methodDeclaration variableDeclaration
%type <object> typedefDeclaration EXTERN_TDeclaration importDeclaration
%type <object> friendDeclaration enumDeclaration
%type <object> operatorDeclaration
%type <object> argumentList argument
%type <object> type

%type <object> commentedDeclaration declaration

%type <string> commentListNoField
%type <object> optComment commentList
%type <object> commentListWithFields commentWithField

%type <i> optQualifiers qualifierList qualifier
%type <i> optIndirects indirectList indirect
%type <i> postDeclaration

%left QUALIFIER
%right '*' '&'

%%

file:
  /* nothing */
  | file commentedDeclaration {
      toplevelObject = Cce_ObjectAppendObject( $2, toplevelObject );
    }
  ;

  /*
    | ignoredToken { $$ = 0; }
    | ';' { $$ = 0; }
  */

  /*
   * Declarations
   */

commentedDeclaration:
    commentList declaration {
      if ($$ = $2) {
	Cce_CMethodWriteComment( $$, $1 );
      } else {
	free($1);
      }
    }
  | declaration
  ;

declaration:
    classDeclaration
  | methodDeclaration
  | operatorDeclaration
  | variableDeclaration
  | typedefDeclaration
  | EXTERN_TDeclaration
  | importDeclaration
  | friendDeclaration
  | enumDeclaration
  | error ';' { $$ = 0; }
  ;

  /*
   * Class Declarations
   */

classDeclaration:
    CLASS IDENTIFIER optBaseList '{' classMemberList '}' ';' {
      $$ = Cce_CClassNew();
      Cce_CClassWriteName( $$, $2 );
      Cce_CClassWriteInherits( $$, $3 );
      Cce_CClassWriteMembers( $$, $5 );
      Cce_MemberListSetClass( $5, $2 );
      SetLocation( $$, @2 );
    }
  | CLASS IDENTIFIER optBaseList '{' error '}' ';' { $$ = 0; }
  | STRUCT IDENTIFIER '{' structMemberList '}' ';' {
      $$ = Cce_CClassNew();
      Cce_CClassWriteName( $$, $2 );
      Cce_CClassWriteMembers( $$, $4 );
      Cce_MemberListSetClass( $4, $2 );
      SetLocation( $$, @2 );
    }
  | STRUCT '{' structMemberList '}' IDENTIFIER ';' {
      $$ = Cce_CClassNew();
      Cce_CClassWriteName( $$, $5 );
      Cce_CClassWriteMembers( $$, $3 );
      Cce_MemberListSetClass( $3, $5 );
      SetLocation( $$, @1 );
    }
  | STRUCT IDENTIFIER '{' error '}' ';' { $$ = 0; }
  | STRUCT IDENTIFIER ';' { $$ = 0; }
  | CLASS IDENTIFIER ';' { $$ = 0; }
  ;

  /*
   * Inheritance information
   */

optBaseList:
    /* nothing */ { $$ = 0; }
  | ':' baseList { $$ = $2; }
  ;

baseList:
    baseSpecifier
  | baseList ',' baseSpecifier { $$ = Cce_ObjectAppendObject( $3, $1 ); }
  ;

baseSpecifier:
    ACCESS IDENTIFIER {
      $$ = Cce_CClassInheritsNew();
      Cce_CClassInheritsWriteName( $$, $2 );
      Cce_CClassInheritsWriteAccess( $$, $1 );
    }
  ;

  /*
   * Class members
   */

classMemberList:  
    optMemberList accessMemberList {
       Cce_MemberListSetAccess( $1, CCE_PRIVATE );
       $$ = Cce_ObjectAppendObject( $2, $1 );
    }
  ;

structMemberList:
  memberList {
      Cce_MemberListSetAccess( $1, CCE_PUBLIC );
      $$ = $1;
    }
  ;

accessMemberList:
    ACCESS ':' optMemberList {
      Cce_MemberListSetAccess( $3, $1 );
      $$ = $3;
    }
  | accessMemberList ACCESS ':' optMemberList {
      Cce_MemberListSetAccess( $4, $2 );
      $$ = Cce_ObjectAppendObject( $4, $1 );
    }
  ;

optMemberList:
    /* nothing */ { $$ = 0; }
  | memberList
  ;

memberList:
    commentedDeclaration
  | memberList commentedDeclaration { $$ = Cce_ObjectAppendObject( $2, $1 ); }
  | memberList error { $$ = $1; }
  ;
 
optComment:
    /* nothing */ { $$ = 0; }
  | commentList
  ;

commentList:
      commentListNoField {
        $$ = Cce_CCommentNew();
	Cce_CCommentWriteField( $$, Cce_StringDup("Synopsis") );
	Cce_CCommentWriteValue( $$, $1 );
      }      			   
    | commentListNoField commentListWithFields {
        Cce_Object c;
	$$ = Cce_CCommentNew();
	Cce_CCommentWriteField( $$, Cce_StringDup("Synopsis") );
	Cce_CCommentWriteValue( $$, $1 );
        $$ = Cce_ObjectAppendObject( $2, $$ );
      }
  ;

commentListNoField:
    COMMENT
  | commentListNoField COMMENT { free($1); $$ = $2; }
  ;

commentListWithFields:
    commentWithField
  | commentListWithFields commentWithField {
      $$ = Cce_ObjectAppendObject( $2, $1 );
    }   
  ;

commentWithField:
    COMMENTFIELD {
      $$ = Cce_CCommentNew();
      Cce_CCommentWriteField( $$, $1 );
     }
  | COMMENTFIELD COMMENT {
      $$ = Cce_CCommentNew();
      Cce_CCommentWriteField( $$, $1 );
      Cce_CCommentWriteValue( $$, $2 );
    }
  ;

optCode:
    ';'
  |  '{' ignoredTokens '}'
  |  '{' ignoredTokens '}' ';'
  ;

  /*
   * Declarations
   */

methodDeclaration:
    type IDENTIFIER '(' argumentList ')' postDeclaration optCode {

      /*
       * Method without a full class name
       */

      $$ = Cce_CMethodNew();
      SetLocation( $$, @2 );
      Cce_CMethodWriteName( $$, $2 );
      Cce_CMethodWriteType( $$, $1 );
      Cce_CMethodWriteArguments( $$, $4 );
      Cce_CMethodWriteQualifiers( $$, $6 );
    }
  | type IDENTIFIER DOUBLECOLON IDENTIFIER '(' argumentList ')' postDeclaration optCode {

      /*
       * Method with a full class name
       */

      $$ = Cce_CMethodNew();
      SetLocation( $$, @2 );
      Cce_CMethodWriteName( $$, $4 );
      Cce_CMethodWriteClassName( $$, $2 );
      Cce_CMethodWriteType( $$, $1 );
      Cce_CMethodWriteArguments( $$, $6 );
      Cce_CMethodWriteQualifiers( $$, $8 );
    }
  | optQualifiers IDENTIFIER '(' argumentList ')' postDeclaration optCode {

      /*
       * Constructor without a full class name
       */

      Cce_Object t;
      t = Cce_CTypeNew();
      Cce_CTypeWriteQualifiers(t, $1);

      $$ = Cce_CMethodNew();
      SetLocation( $$, @2 );
      Cce_CMethodWriteType( $$, t );
      Cce_CMethodWriteName( $$, $2 );
      Cce_CMethodWriteArguments( $$, $4 );
      Cce_CMethodWriteQualifiers( $$, $6 );
    }
  | optQualifiers IDENTIFIER DOUBLECOLON IDENTIFIER '(' argumentList ')' postDeclaration optCode {

      /*
       * Constructor with a full class name
       */

      Cce_Object t;
      t = Cce_CTypeNew();
      Cce_CTypeWriteQualifiers(t, $1);

      $$ = Cce_CMethodNew();
      SetLocation( $$, @2 );
      Cce_CMethodWriteType( $$, t );
      Cce_CMethodWriteName( $$, $4 );
      Cce_CMethodWriteClassName( $$, $2 );
      Cce_CMethodWriteArguments( $$, $6 );
      Cce_CMethodWriteQualifiers( $$, $8 );
    }
  | optQualifiers '~' IDENTIFIER '(' ')' optCode {

      /*
       * Destructor without a full class name
       */

      Cce_Object t;
      char * name;
      t = Cce_CTypeNew();
      Cce_CTypeWriteQualifiers(t, $1);

      $$ = Cce_CMethodNew();
      SetLocation( $$, @2 );
      name = (char *) malloc(strlen($3)+2);
      name[0] = '~';
      strcpy(name+1,$3);
      Cce_CMethodWriteName( $$, name );
      Cce_CMethodWriteType( $$, t );
    }
  | optQualifiers IDENTIFIER DOUBLECOLON '~' IDENTIFIER '(' ')' optCode {

      /*
       * Destructor with a full class name
       */

      Cce_Object t;
      char * name;

      t = Cce_CTypeNew();
      Cce_CTypeWriteQualifiers(t, $1);

      $$ = Cce_CMethodNew();
      SetLocation( $$, @2 );
      name = (char *) malloc(strlen($5)+2);
      name[0] = '~';
      strcpy(name+1,$5);
      Cce_CMethodWriteName( $$, name );
      Cce_CMethodWriteClassName( $$, $2 );
      Cce_CMethodWriteType( $$, t );
    }
  ;

variableDeclaration:
    type IDENTIFIER optInitialization {

      /*
       * Scalar variable without a full class name
       */

      $$ = Cce_CVariableNew();
      SetLocation( $$, @2 );
      Cce_CVariableWriteName( $$, $2 );
      Cce_CVariableWriteType( $$, $1 );
    }
  | type IDENTIFIER DOUBLECOLON IDENTIFIER optInitialization {

      /*
       * Scalar variable without a full class name
       */

      $$ = Cce_CVariableNew();
      SetLocation( $$, @2 );
      Cce_CVariableWriteName( $$, $4 );
      Cce_CVariableWriteClassName( $$, $2 );
      Cce_CVariableWriteType( $$, $1 );
    }
  | type IDENTIFIER '[' ignoredTokens ']' optInitialization {

      /*
       * Array variable without a full class name
       */

      $$ = Cce_CVariableNew();
      SetLocation( $$, @2 );
      Cce_CVariableWriteName( $$, $2 );
      Cce_CVariableWriteType( $$, $1 );
      Cce_CTypeWriteQualifiers( $1, Cce_CTypeReadQualifiers( $1 )
				      | CCE_ARRAY );
   }
  | type IDENTIFIER DOUBLECOLON IDENTIFIER '[' ignoredTokens ']' optInitialization {

      /*
       * Array variable with a full class name
       */

      $$ = Cce_CVariableNew();
      SetLocation( $$, @2 );
      Cce_CVariableWriteName( $$, $4 );
      Cce_CVariableWriteClassName( $$, $2 );
      Cce_CVariableWriteType( $$, $1 );
      Cce_CTypeWriteQualifiers( $1, Cce_CTypeReadQualifiers( $1 )
				      | CCE_ARRAY );
   }
  ;

optInitialization:
   ';'
 | '=' ignoredUntilSemicolon ';'
 ;

operatorDeclaration:
    type OPERATOR operator '(' argumentList ')' postDeclaration optCode {

      /*
       * Operator without a full class name	
       */

      $$ = Cce_COperatorNew();
      SetLocation( $$, @2 );
      Cce_COperatorWriteOperator( $$, $3 );
      Cce_COperatorWriteType( $$, $1 );
      Cce_COperatorWriteArguments( $$, $5 );
      Cce_COperatorWriteQualifiers( $$, $7 );
    }
  | type IDENTIFIER DOUBLECOLON OPERATOR operator '(' argumentList ')' postDeclaration optCode {

      /*
       * Operator with a full class name
       */

      $$ = Cce_COperatorNew();
      SetLocation( $$, @2 );
      Cce_COperatorWriteOperator( $$, $5 );
      Cce_COperatorWriteClassName( $$, $2 );
      Cce_COperatorWriteType( $$, $1 );
      Cce_COperatorWriteArguments( $$, $7 );
      Cce_COperatorWriteQualifiers( $$, $9 );
    }
  | optQualifiers OPERATOR type '(' argumentList ')' postDeclaration optCode {

      /*
       * Casting operator without a full class name
       */

      $$ = Cce_COperatorNew();
      SetLocation( $$, @2 );
      Cce_COperatorWriteOperator( $$, Cce_StringDup("cast") );
      Cce_COperatorWriteType( $$, $3 );
      Cce_COperatorWriteArguments( $$, $5 );
      Cce_COperatorWriteQualifiers( $$, $1 | $7 );
    }
  | optQualifiers IDENTIFIER DOUBLECOLON OPERATOR type '(' argumentList ')' postDeclaration optCode {

      /*
       * Casting operator with a full class name
       */

      $$ = Cce_COperatorNew();
      SetLocation( $$, @2 );
      Cce_COperatorWriteOperator( $$, Cce_StringDup("cast") );
      Cce_COperatorWriteType( $$, $5 );
      Cce_COperatorWriteArguments( $$, $7 );
      Cce_COperatorWriteQualifiers( $$, $1 | $9 );
    }
  ;

importDeclaration:
    optQualifiers IDENTIFIER DOUBLECOLON IDENTIFIER ';' {
       free($2); free($4);
    }
  ;

friendDeclaration:
    FRIEND ignoredUntilSemicolon ';' { $$ = 0; }
  ;

enumDeclaration:
     ENUM optIdentifier '{' ignoredTokens '}' ';' { if ($2) free($2); $$ = 0; }
  |  UNION optIdentifier '{' ignoredTokens '}' ';' { if ($2) free($2); $$ = 0; }
  ;

optIdentifier:
    /* nothing */ { $$ = 0; }
  | IDENTIFIER
  ;

typedefDeclaration:
    TYPEDEF ignoredUntilSemicolon ';' { $$ = 0; }
  ;

EXTERN_TDeclaration:
    EXTERN_T ignoredUntilSemicolon ';' { $$ = 0; }
  ;

/*
 * Types
 */

type:
    optQualifiers IDENTIFIER optIndirects {
       $$ = Cce_CTypeNew();
       Cce_CTypeWriteName( $$, $2 );
       Cce_CTypeWriteQualifiers( $$, $1 );
       Cce_CTypeWriteIndirect( $$, $3 );
    }
  ;

optIndirects:
    /* nothing */ { $$ = 0; }
  | indirectList
  ;

indirectList:
    indirect
  | indirectList indirect{ $$ = $1 + $2; }
  ;

indirect:
    '*' { $$ = 1; }
  | '&' { $$ = -1; }
  ;

optQualifiers:
    /* nothing */ { $$ = 0; }
  | qualifierList
  ;

qualifierList:
    qualifier
  | qualifierList qualifier { $$ = $1 | $2; }
  ;

qualifier:
    QUALIFIER
  | '=' '0' { $$ = CCE_PURE; }
  ;

postDeclaration:
    /* nothing */ { $$ = 0; }
  | qualifierList
  | qualifierList ':' initializers { $$ = $1; }
  | ':' initializers { $$ = 0; }
  ;

initializers:
    initializer
  | initializers ',' initializer
  ;

initializer:
    IDENTIFIER '(' ignoredTokens ')' { free($1); }
  ;

argumentList:
    /* nothing */ { $$ = 0; }
  | argument optComment { $$ = $1;  Cce_CArgumentWriteComment( $$, $2 ); }
  | argumentList ',' argument optComment {
      $$ = Cce_ObjectAppendObject( $3, $1 );
      Cce_CArgumentWriteComment( $$, $4 );
    }
  ;

argument:
    type optIdentifier optDefault {
      $$ = Cce_CArgumentNew();
      Cce_CArgumentWriteName( $$, $2 );
      Cce_CArgumentWriteType( $$, $1 );
      Cce_CArgumentWriteDefault( $$, $3 );
    }
  | type optIdentifier '[' ignoredTokens ']' optDefault {
      $$ = Cce_CArgumentNew();
      Cce_CArgumentWriteName( $$, $2 );
      Cce_CArgumentWriteType( $$, $1 );
      Cce_CTypeWriteQualifiers( $1, Cce_CTypeReadQualifiers( $1 )
				| CCE_ARRAY );				
      Cce_CArgumentWriteDefault( $$, $6 );
    }
  ;

optDefault:
    /* nothing */ { $$ = 0; }
  | '=' defaultValue { $$ = $2; }
  ;

defaultValue:
    IDENTIFIER
  | TOKEN
  | '0' { $$ = Cce_StringDup("0"); }
  ;

  /* 
   * Operators
   */

operator:
    OP
  | '*' { $$ = Cce_StringDup("*"); }
  | '&' { $$ = Cce_StringDup("&"); }
  | ',' { $$ = Cce_StringDup(","); }
  | '=' { $$ = Cce_StringDup("="); }
  | '~' { $$ = Cce_StringDup("~"); }
  | '[' ']' { $$ = Cce_StringDup("[]"); }
  ;

  /*
   * Ignored things
   */
  
ignoredToken:
    FRIEND
  | UNION
  | ENUM
  | OPERATOR
  | TYPEDEF
  | EXTERN_T
  | ACCESS {}
  | QUALIFIER {}
  | DOUBLECOLON
  | ',' | ':' | '*' | '&' | '~' | '=' | '0'
  | OP { free($1); }
  | IDENTIFIER { free($1); }
  | TOKEN { free($1); }
  | '{' ignoredTokens '}'  
  | '(' ignoredTokens ')'
  | '[' ignoredTokens ']'
  ;

ignoredTokens:
    /* nothing */
  | ignoredTokenList
  ;

ignoredUntilSemicolon:
    /* nothing */
  | ignoredUntilSemicolon ignoredToken
  | ignoredUntilSemicolon CLASS
  | ignoredUntilSemicolon STRUCT
  | ignoredUntilSemicolon COMMENT { free($2); }
  | ignoredUntilSemicolon COMMENTFIELD { free($2); }
  ;

ignoredTokenList:
    braceToken
  | braceToken ignoredTokenList
  ;

braceToken:
    CLASS
  | STRUCT
  | COMMENT { free($1); }
  | COMMENTFIELD { free($1); }
  | ';'
  | ignoredToken
  ;

%%

/**Function********************************************************************
  Synopsis           [ Set the access on a list of members ]
  SideEffects        []
******************************************************************************/
void
Cce_MemberListSetAccess( Cce_Object list, int access )
{
  while ( list != 0 ) {
    Cce_CMethodWriteAccess( list, access );
    list = Cce_ObjectReadNext( list );
  }
}

/**Function********************************************************************
  Synopsis           [ Set the class name on a list of members ]
  SideEffects        []
******************************************************************************/
void
Cce_MemberListSetClass( Cce_Object list, char * className )
{
  while ( list != 0 ) {
    Cce_CMethodWriteClassName( list, Cce_StringDup( className ) );
    list = Cce_ObjectReadNext( list );
  }
}


/**Function********************************************************************

  Synopsis           [ Parse a file ]

  Description	     [ Parse a file and return the status (0 if no problem).
		       If object is non-NIL, fill it in with the top level
		       object from the parse tree.  If parseErrors is non-NIL,
		       append any parse errors to that dynamic string. ]
  SideEffects        []
******************************************************************************/
int
Cce_ParseFile( FILE * file, char * filename, Cce_Object * object,
	       Tcl_DString * parseErrors)
{

  status = 0;
  ccelineno = 1;
  toplevelObject = 0;

  currentFilename = filename;
  parseErrorString = parseErrors;

  ccerestart(file);
  cceparse();

  if ( object ) {
    *object = toplevelObject;
  }

  return status;
}

cceerror(char * msg)
{
  char buf[20];

  if ( parseErrorString ) {
    
    /*
     * Append "{message token lineNumber}" to the error result
     */

    Tcl_DStringStartSublist( parseErrorString );
    Tcl_DStringAppendElement( parseErrorString, msg );
    Tcl_DStringAppendElement( parseErrorString, ccetext );
    sprintf(buf, "%d", ccelineno);    
    Tcl_DStringAppendElement( parseErrorString, buf );
    Tcl_DStringEndSublist( parseErrorString );
  } else {
    printf("%d: %s at '%s'\n", ccelineno, msg, ccetext);
  }
  status = ccelineno;
}
