/**CFile***********************************************************************

  FileName    [ cceObjects.c ]

  PackageName [ cce ]

  Synopsis    [ Functions for cce objects. ]

  Description []

  Author      [ Stephen A. Edwards <sedwards@eecs.berkeley.edu> ]

  Copyright   []

******************************************************************************/

#include "cceInt.h"
#include <string.h>

/**Variable********************************************************************
  Synopsis    [ An array of strings naming the possible slot types. ]
******************************************************************************/
char * cceSlotTypeNames[] = { "objects", "string", "int" };

/**Variable********************************************************************
  Synopsis    [ An array of pointers to all objects in the system,
		indexed by ID ]

  Description [ This array grows as necessary.  cce_numObjects is the
		current number of objects, objectTableSize is the size
		of the memory allocated for it.
		]

******************************************************************************/
Cce_Object * cce_objectPointers;
int cce_numObjects;
static int objectTableSize;

Cce_Object cce_classClass;
Cce_Object cce_slotClass;
#include "cceObjectVariables.include"

/**AutomaticStart*************************************************************/

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/


/**AutomaticEnd***************************************************************/



/**Function********************************************************************

  Synopsis           [ Create a new object of a given class. ]

  SideEffects        [ Adds the new object's ID to the array of
		       object identifier pointers. ]

  SeeAlso            []

******************************************************************************/
Cce_Object
Cce_ObjectNew( Cce_Object class )
{
  Cce_Object newObject;

  /*
   * Create the new object and initialize its class, Id, and next slots
   */

  newObject = (Cce_Object) malloc( sizeof(void *) *
				   Cce_ClassReadNumSlots(class) );
  Cce_ObjectWriteClass( newObject, class );
  Cce_ObjectWriteId( newObject, cce_numObjects );
  Cce_ObjectWriteNext( newObject, 0 );

  /*
   * Zero the remaining slots
   */

  memset( &newObject[3], 0,
	  sizeof(void *) * (Cce_ClassReadNumSlots(class) - 3) );


  /*
   * Update the object counts
   */

  cce_numObjects++;
  ((int) class[ CCE_CLASS_NUM_OBJECTS ])++;

  /*
   * Register the new object
   */
  
  CceRegisterObject( newObject );

  return newObject;
}

/**Function********************************************************************

  Synopsis           [ Create a new slot of a class ]

  Description        [ Create a new object of class Slot and append it
  		        to the class's list of slots.  This should
			<EM>not</EM> be called once objects of the class
			have been created.  Return the slot number.
			The name is copied.
			]

  SideEffects        [ ]

  SeeAlso            [ Cce_ClassNew ]

******************************************************************************/
int
Cce_ClassAddSlot( Cce_Object class, char * name, int type )
{

  Cce_Object newSlot;

  newSlot = Cce_ObjectNew( cce_slotClass );
  Cce_ObjectSlotWrite( newSlot, CCE_SLOT_NAME, Cce_StringDup(name) );
  Cce_ObjectSlotWrite( newSlot, CCE_SLOT_TYPE, type );

  Cce_ObjectAppendObject( newSlot, (Cce_Object) class[ CCE_CLASS_SLOTS ] );
  return ((int) class[ CCE_CLASS_NUM_SLOTS ])++;
}


/**Function********************************************************************

  Synopsis           [ Create a new class ]

  Description        [ Create a new object of class Class and fill in
		       the first few slots ]

  SideEffects        [ ]

  SeeAlso	     [ Cce_ObjectNew ]

******************************************************************************/
Cce_Object
Cce_ClassNew(char * name)
{
  Cce_Object newClass;
  Cce_Object classSlot;

  /*
   * Create the new class object and append it to the list of all classes
   */

  newClass = Cce_ObjectNew( cce_classClass );
  Cce_ObjectAppendObject( newClass, cce_classClass );
  Cce_ObjectSlotWrite( newClass, CCE_CLASS_NAME, Cce_StringDup(name) );
  Cce_ObjectSlotWrite( newClass, CCE_CLASS_NUM_SLOTS, 1 );

  /*
   * Create the "class" slot and add it in
   */

  classSlot = Cce_ObjectNew( cce_slotClass );
  Cce_ObjectSlotWrite( classSlot, CCE_SLOT_NAME, Cce_StringDup("class") );
  Cce_ObjectSlotWrite( classSlot, CCE_SLOT_TYPE, CCE_OBJECT );
  Cce_ObjectSlotWrite( newClass, CCE_CLASS_SLOTS, classSlot );

  /*
   * Add the "id" and "next" slots
   */

  Cce_ClassAddSlot( newClass, "id", CCE_INTEGER );
  Cce_ClassAddSlot( newClass, "next", CCE_OBJECT );

  return newClass;
}


/**Function********************************************************************

  Synopsis           [ Register an object in the global array ]

  Description        [ First, resize the array if necessary, then register the
		       object. ]

  SideEffects        [ cce_numObjects, objectTableSize, and cce_objectPointers
  		       may be modified ]

******************************************************************************/
void
CceRegisterObject( Cce_Object object )
{
  int id = Cce_ObjectReadId( object );
  Cce_Object * newObjectPointers;
  int newTableSize;

  if ( id >= objectTableSize ) {
    
    /*
     * Too many objects--resize the object table
     */

    newTableSize = 2 * objectTableSize;
    newObjectPointers = (Cce_Object *) malloc( sizeof(Cce_Object) * newTableSize );
    memcpy( newObjectPointers, cce_objectPointers,
	    sizeof(Cce_Object) * objectTableSize );
    free( cce_objectPointers );
    cce_objectPointers = newObjectPointers;
    objectTableSize = newTableSize;        
  }

  cce_objectPointers[id] = object;

}

/**Function********************************************************************

  Synopsis           [ Initialize the object system ]

  Description        []

  SideEffects        [ The object pointer table is initialized ]

  SeeAlso            [ ]

******************************************************************************/
void
Cce_ObjectsInitialize()
{

  Cce_Object classSlot;

  /*
   * Clear the object table
   */

  cce_objectPointers = (Cce_Object *) malloc( sizeof(Cce_Object) * 32 );
  objectTableSize = 32;

  /*
   * Build the object for the Class class
   */

  cce_classClass = (Cce_Object) malloc( sizeof(void *) * 7 );
  Cce_ObjectWriteClass( cce_classClass, cce_classClass );
  Cce_ObjectWriteId( cce_classClass, 0 );
  Cce_ObjectWriteNext( cce_classClass, 0 );
  Cce_ObjectSlotWrite( cce_classClass, CCE_CLASS_NAME, Cce_StringDup("Class") );
  Cce_ObjectSlotWrite( cce_classClass, CCE_CLASS_NUM_OBJECTS, 1 );
  Cce_ObjectSlotWrite( cce_classClass, CCE_CLASS_NUM_SLOTS, 7 );

  cce_objectPointers[0] = cce_classClass;
  cce_numObjects = 1;

  /*
   * Build the object for the Slot class
   */

  cce_slotClass = Cce_ObjectNew( cce_classClass );
  Cce_ObjectAppendObject( cce_slotClass, cce_classClass );
  Cce_ObjectSlotWrite( cce_slotClass, CCE_CLASS_NAME, Cce_StringDup("Slot") );
  Cce_ObjectSlotWrite( cce_slotClass, CCE_CLASS_NUM_OBJECTS, 0 );
  Cce_ObjectSlotWrite( cce_slotClass, CCE_CLASS_NUM_SLOTS, 5 );

  /*
   * Add the slots to the Class class
   *
   * A trick--Cce_ClassAddSlot modified the number of slots, so
   * we correct it after all slots are added
   */

  classSlot = Cce_ObjectNew( cce_slotClass );
  Cce_ObjectSlotWrite( classSlot, CCE_SLOT_NAME, Cce_StringDup("class") );
  Cce_ObjectSlotWrite( classSlot, CCE_SLOT_TYPE, CCE_OBJECT );
  Cce_ObjectSlotWrite( cce_classClass, CCE_CLASS_SLOTS, classSlot );

  Cce_ClassAddSlot( cce_classClass, "id", CCE_INTEGER );
  Cce_ClassAddSlot( cce_classClass, "next", CCE_OBJECT );
  Cce_ClassAddSlot( cce_classClass, "name", CCE_STRING );
  Cce_ClassAddSlot( cce_classClass, "numObjects", CCE_INTEGER );
  Cce_ClassAddSlot( cce_classClass, "numSlots", CCE_INTEGER );
  Cce_ClassAddSlot( cce_classClass, "slots", CCE_OBJECT );

  Cce_ObjectSlotWrite( cce_classClass, CCE_CLASS_NUM_SLOTS, 7 );

  /*
   * Add the slots to the Slot class
   *
   * An even nastier trick--we're adding slots to Slot,
   * which changes the number of slots it thinks it has.  We correct
   * the count after each call.
   */

  classSlot = Cce_ObjectNew( cce_slotClass );
  Cce_ObjectSlotWrite( classSlot, CCE_SLOT_NAME, Cce_StringDup("class") );
  Cce_ObjectSlotWrite( classSlot, CCE_SLOT_TYPE, CCE_OBJECT );
  Cce_ObjectSlotWrite( cce_slotClass, CCE_CLASS_SLOTS, classSlot );

  Cce_ClassAddSlot( cce_slotClass, "id", CCE_INTEGER );
  Cce_ObjectSlotWrite( cce_slotClass, CCE_CLASS_NUM_SLOTS, 5 );
  Cce_ClassAddSlot( cce_slotClass, "next", CCE_OBJECT );
  Cce_ObjectSlotWrite( cce_slotClass, CCE_CLASS_NUM_SLOTS, 5 );
  Cce_ClassAddSlot( cce_slotClass, "name", CCE_STRING );
  Cce_ObjectSlotWrite( cce_slotClass, CCE_CLASS_NUM_SLOTS, 5 );
  Cce_ClassAddSlot( cce_slotClass, "type", CCE_INTEGER );
  Cce_ObjectSlotWrite( cce_slotClass, CCE_CLASS_NUM_SLOTS, 5 );

  /*
   * Construct the built-in classes
   */

#include "cceObjectCreators.include"

}

/**Function********************************************************************

  Synopsis           [ Append an object to the list ]

  Description	     [ Append an object to the list and return the list.
		       If the list is NULL, return the object.
		       If the object is NULL, return the list. ]

  SideEffects        []

******************************************************************************/
Cce_Object
Cce_ObjectAppendObject( Cce_Object object, Cce_Object list )
{
  Cce_Object lastObject;
  Cce_Object nextObject;

  if ( list == 0 ) {
    return object;
  }

  if ( object == 0 ) {
    return list;
  }

  /*
   * Find the end of the list
   */

  lastObject = list;
  nextObject = Cce_ObjectReadNext( list );
  while ( nextObject ) {
    lastObject = nextObject;
    nextObject = Cce_ObjectReadNext( nextObject );
  }

  Cce_ObjectWriteNext( lastObject, object );
  return list;
}

/**Function********************************************************************

  Synopsis           [ Return the index of a slot name ]

  Description        [ Return the index of a slot name in the given class,
		       or -1 if no slot was found.  Also fills in the
		       slot type if the pointer is non-nil. ]

  SideEffects        []

******************************************************************************/
int
Cce_ClassFindNamedSlot( Cce_Object class, char * slotName, int * type )
{
  int i = 0;
  Cce_Object slot;

  slot = Cce_ClassReadSlots( class );

  while ( slot != 0 ) {
    if ( strcmp( Cce_SlotReadName( slot ), slotName ) == 0 ) {
      break;
    }
    slot = Cce_ObjectReadNext( slot );
    i++;
  }

  if ( slot ) {
    if ( type ) {
      *type = Cce_SlotReadType( slot );
    }
    return i;
  } else {
    return -1;
  }  

}


